@@ -8,6 +8,8 @@
 #include <TNL/Allocators/CudaHost.h>
 #include <TNL/Allocators/CudaManaged.h>
 
+#include "dlpack.h"
+
 template< typename Index >
 void
 ndarray_check_index( std::size_t i, Index idx, Index size )
@@ -49,7 +51,10 @@ ndarray_indexing( nb::class_< ArrayType, Args... >& array )
          std::array< IndexType, dim > indices_array;
          for( std::size_t i = 0; i < dim; ++i ) {
             indices_array[ i ] = nb::cast< IndexType >( tuple_indices[ i ] );
-            ndarray_check_index( i, indices_array[ i ], self.getSizes()[ i ] );
+            if( ArrayType::SizesHolderType::getStaticSize( i ) > 0 )
+               ndarray_check_index( i, indices_array[ i ], ArrayType::SizesHolderType::getStaticSize( i ) );
+            else
+               ndarray_check_index( i, indices_array[ i ], self.getSizes()[ i ] );
          }
 
          // Unpack the array into the operator()
@@ -67,34 +72,41 @@ ndarray_indexing( nb::class_< ArrayType, Args... >& array )
       "__setitem__",
       [ dim ]( ArrayType& self, nb::object indices, ValueType value )
       {
-         nb::tuple tuple_indices;
-
-         if( nb::isinstance< nb::tuple >( indices ) ) {
-            tuple_indices = nb::cast< nb::tuple >( indices );
-         }
+         if constexpr( std::is_const_v< ValueType > )
+            throw nb::type_error( "Cannot set element of a read-only array" );
          else {
-            tuple_indices = nb::make_tuple( indices );
-         }
+            nb::tuple tuple_indices;
 
-         if( tuple_indices.size() != dim ) {
-            throw nb::value_error( ( "Expected " + std::to_string( dim ) + " indices" ).c_str() );
-         }
+            if( nb::isinstance< nb::tuple >( indices ) ) {
+               tuple_indices = nb::cast< nb::tuple >( indices );
+            }
+            else {
+               tuple_indices = nb::make_tuple( indices );
+            }
 
-         std::array< IndexType, dim > indices_array;
-         for( std::size_t i = 0; i < dim; ++i ) {
-            indices_array[ i ] = nb::cast< IndexType >( tuple_indices[ i ] );
-            ndarray_check_index( i, indices_array[ i ], self.getSizes()[ i ] );
-         }
+            if( tuple_indices.size() != dim ) {
+               throw nb::value_error( ( "Expected " + std::to_string( dim ) + " indices" ).c_str() );
+            }
 
-         // Unpack and assign
-         std::apply(
-            [ & ]( auto... indices )
-            {
-               // setElement is equivalent to operator[] on host but works on cuda
-               const auto idx = self.getStorageIndex( indices... );
-               self.getStorageArray().setElement( idx, value );
-            },
-            indices_array );
+            std::array< IndexType, dim > indices_array;
+            for( std::size_t i = 0; i < dim; ++i ) {
+               indices_array[ i ] = nb::cast< IndexType >( tuple_indices[ i ] );
+               if( ArrayType::SizesHolderType::getStaticSize( i ) > 0 )
+                  ndarray_check_index( i, indices_array[ i ], ArrayType::SizesHolderType::getStaticSize( i ) );
+               else
+                  ndarray_check_index( i, indices_array[ i ], self.getSizes()[ i ] );
+            }
+
+            // Unpack and assign
+            std::apply(
+               [ & ]( auto... indices )
+               {
+                  // setElement is equivalent to operator[] on host but works on cuda
+                  const auto idx = self.getStorageIndex( indices... );
+                  self.getStorageArrayView().setElement( idx, value );
+               },
+               indices_array );
+         }
       },
       nb::arg( "indices" ),
       nb::arg( "value" ) );
@@ -286,7 +298,6 @@ export_NDArray( nb::module_& m, const char* name )
    using IndexerType = typename ArrayType::IndexerType;
    using ValueType = typename ArrayType::ValueType;
    using DeviceType = typename ArrayType::DeviceType;
-   using AllocatorType = typename ArrayType::AllocatorType;
 
    auto array =  //
       nb::class_< ArrayType, IndexerType >( m, name )
@@ -319,11 +330,154 @@ export_NDArray( nb::module_& m, const char* name )
                   return nb::type< ValueType >();
                }
             } )
+         .def_prop_ro_static(  //
+            "ViewType",
+            []( nb::handle ) -> nb::typed< nb::handle, nb::type_object >
+            {
+               return nb::type< typename ArrayType::ViewType >();
+            } )
+         .def_prop_ro_static(  //
+            "ConstViewType",
+            []( nb::handle ) -> nb::typed< nb::handle, nb::type_object >
+            {
+               return nb::type< typename ArrayType::ConstViewType >();
+            } )
 
          // Constructors
          .def( nb::init<>() )
          .def( nb::init< ArrayType >(), nb::arg( "other" ) )
 
+         // NDArrayView getters
+         .def( "getView", &ArrayType::getView )
+         .def( "getConstView", &ArrayType::getConstView )
+         // TODO: getSubarrayView (requires template parameters...)
+
+         // Internal storage
+         .def( "getStorageArrayView",
+               &ArrayType::getStorageArrayView,
+               nb::rv_policy::reference_internal,
+               "Return an ArrayView for the underlying storage array." )
+
+         // Assignment
+         .def( "assign",
+               []( ArrayType& array, const ArrayType& other ) -> ArrayType&
+               {
+                  if constexpr( std::is_const_v< ValueType > )
+                     throw nb::type_error( "Cannot assign into constant array" );
+                  else
+                     return array = other;
+               } )
+
+         // Comparison
+         .def( nb::self == nb::self, nb::sig( "def __eq__(self, arg: object, /) -> bool" ) )
+         .def( nb::self != nb::self, nb::sig( "def __ne__(self, arg: object, /) -> bool" ) )
+
+         // String representation
+         .def(
+            "__str__",
+            []( const ArrayType& self )
+            {
+               constexpr std::size_t dim = ArrayType::getDimension();
+               std::ostringstream oss;
+               oss << "NDArray[" << dim << ", ";
+               if constexpr( std::is_same_v< ValueType, bool > ) {
+                  oss << "bool";
+               }
+               else if constexpr( std::is_integral_v< ValueType > ) {
+                  oss << "int";
+               }
+               else if constexpr( std::is_floating_point_v< ValueType > ) {
+                  oss << "float";
+               }
+               else if constexpr( TNL::is_complex_v< ValueType > ) {
+                  oss << "complex";
+               }
+               else {
+                  oss << TNL::getType( ValueType{} );
+               }
+               oss << ", ";
+               if constexpr( std::is_same_v< DeviceType, TNL::Devices::Cuda > )
+                  oss << "Cuda";
+               else
+                  oss << "Host";
+               oss << "](";
+               TNL::Algorithms::staticFor< std::size_t, 0, dim >(
+                  [ & ]( auto i )
+                  {
+                     if constexpr( i > 0 )
+                        oss << ", ";
+                     oss << self.template getSize< i >();
+                  } );
+               oss << ")";
+               return oss.str();
+            },
+            "Returns a readable string representation of the array" )
+      //
+      ;
+
+   // Interoperability with Python array API standard (DLPack)
+   // (note that the set of dtypes supported by DLPack is limited)
+   if constexpr( nb::dtype< ValueType >().bits != 0 ) {
+      array
+         .def(
+            "__dlpack__",
+            []( nb::pointer_and_handle< ArrayType > self, nb::kwargs kwargs )
+            {
+               int device_id = 0;
+               // FIXME: DLPack support switching CUDA devices but TNL does not
+               if constexpr( std::is_same_v< DeviceType, TNL::Devices::Cuda > )
+                  device_id = TNL::Backend::getDevice();
+
+               constexpr std::size_t dim = ArrayType::getDimension();
+               std::array< std::size_t, dim > sizes;
+               std::array< std::int64_t, dim > strides;
+               TNL::Algorithms::staticFor< std::size_t, 0, dim >(
+                  [ & ]( auto i )
+                  {
+                     sizes[ i ] = self.p->template getSize< i >();
+                     strides[ i ] = self.p->template getStride< i >();
+                  } );
+
+               using array_api_t = nb::ndarray< nb::array_api, ValueType >;
+               array_api_t array_api( self.p->getData(),
+                                      dim,
+                                      sizes.data(),
+                                      self.h,  // pass the Python object associated with `self` as owner
+                                      strides.data(),
+                                      nb::dtype< ValueType >(),
+                                      dlpack_device< ArrayType >().first,
+                                      device_id );
+
+               // call the array_api's __dlpack__ Python method to properly handle the kwargs
+               nb::object aa = nb::cast( array_api, nb::rv_policy::reference_internal, self.h );
+               return aa.attr( "__dlpack__" )( **kwargs );
+            },
+            nb::sig( "def __dlpack__(self, **kwargs: typing.Any) -> typing_extensions.CapsuleType" ) )
+         .def_static( "__dlpack_device__", dlpack_device< ArrayType > );
+   }
+
+   ndarray_indexing( array );
+   ndarray_iteration( array );
+
+   if constexpr( TNL::IsViewType< ArrayType >::value ) {
+      array  //
+         .def( nb::init< const ArrayType& >() )
+         // FIXME: needed for implicit conversion from NDArray, but AllocatorType is ignored
+         //.def( nb::init_implicit< TNL::Containers::NDArray< ... >& >() )
+         .def( "bind",
+               []( ArrayType& self, const ArrayType& other )
+               {
+                  self.bind( other );
+               } )
+         .def( "reset",
+               &ArrayType::reset,
+               "Reset the array view to the empty state. There is no deallocation, it does not affect other views." )
+         //
+         ;
+   }
+   else {
+      // Additional NDArray-specific methods
+      array
          // Size management
          .def( "setSizes",
                &ArrayType::setSize,  // Note: C++ method should be renamed
@@ -367,67 +521,9 @@ export_NDArray( nb::module_& m, const char* name )
                "Reset the array to the empty state. The current data will be deallocated, "
                "thus all pointers and views to the array elements will become invalid." )
 
-         // Assignment
-         .def( "assign",
-               []( ArrayType& array, const ArrayType& other ) -> ArrayType&
-               {
-                  return array = other;
-               } )
-
-         // Comparison
-         .def( nb::self == nb::self, nb::sig( "def __eq__(self, arg: object, /) -> bool" ) )
-         .def( nb::self != nb::self, nb::sig( "def __ne__(self, arg: object, /) -> bool" ) )
-
          // Fill
          .def( "setValue", &ArrayType::setValue, nb::arg( "value" ) )
 
-         // Internal storage
-         .def( "getStorageArray",
-               nb::overload_cast<>( &ArrayType::getStorageArray ),
-               nb::rv_policy::reference_internal,
-               "Return a reference to the underlying storage array." )
-
-         // String representation
-         .def(
-            "__str__",
-            []( const ArrayType& self )
-            {
-               constexpr std::size_t dim = ArrayType::getDimension();
-               std::ostringstream oss;
-               oss << "NDArray[" << dim << ", ";
-               if constexpr( std::is_same_v< ValueType, bool > ) {
-                  oss << "bool";
-               }
-               else if constexpr( std::is_integral_v< ValueType > ) {
-                  oss << "int";
-               }
-               else if constexpr( std::is_floating_point_v< ValueType > ) {
-                  oss << "float";
-               }
-               else if constexpr( TNL::is_complex_v< ValueType > ) {
-                  oss << "complex";
-               }
-               else {
-                  oss << TNL::getType( ValueType{} );
-               }
-               oss << ", ";
-               if constexpr( std::is_same_v< DeviceType, TNL::Devices::Cuda > )
-                  oss << "Cuda";
-               else
-                  oss << "Host";
-               oss << "](";
-               TNL::Algorithms::staticFor< std::size_t, 0, dim >(
-                  [ & ]( auto i )
-                  {
-                     if constexpr( i > 0 )
-                        oss << ", ";
-                     oss << self.template getSize< i >();
-                  } );
-               oss << ")";
-               return oss.str();
-            },
-            "Returns a readable string representation of the array" )
-
          // Deepcopy support https://pybind11.readthedocs.io/en/stable/advanced/classes.html#deepcopy-support
          .def( "__copy__",
                []( const ArrayType& self )
@@ -441,83 +537,7 @@ export_NDArray( nb::module_& m, const char* name )
                return ArrayType( self );
             },
             nb::arg( "memo" ) );
-
-   // Interoperability with Python array API standard (DLPack)
-   auto dlpack_device = []()
-   {
-      // FIXME: DLPack supports switching CUDA devices but TNL does not
-      if constexpr( std::is_same_v< AllocatorType, TNL::Allocators::Cuda< ValueType > > )
-         return std::make_pair( nb::device::cuda::value, TNL::Backend::getDevice() );
-      else if constexpr( std::is_same_v< AllocatorType, TNL::Allocators::CudaHost< ValueType > > )
-         return std::make_pair( nb::device::cuda_host::value, TNL::Backend::getDevice() );
-      else if constexpr( std::is_same_v< AllocatorType, TNL::Allocators::CudaManaged< ValueType > > )
-         return std::make_pair( nb::device::cuda_managed::value, TNL::Backend::getDevice() );
-      else
-         return std::make_pair( nb::device::cpu::value, 0 );
-   };
-   array
-      .def(
-         "__dlpack__",
-         [ dlpack_device ]( ArrayType& self, nb::kwargs kwargs )
-         {
-            int device_id = 0;
-            // FIXME: DLPack support switching CUDA devices but TNL does not
-            if constexpr( std::is_same_v< DeviceType, TNL::Devices::Cuda > )
-               device_id = TNL::Backend::getDevice();
-
-            constexpr std::size_t dim = ArrayType::getDimension();
-            std::array< std::size_t, dim > sizes;
-            std::array< std::int64_t, dim > strides;
-            TNL::Algorithms::staticFor< std::size_t, 0, dim >(
-               [ & ]( auto i )
-               {
-                  sizes[ i ] = self.template getSize< i >();
-                  strides[ i ] = self.template getStride< i >();
-               } );
-
-            return nb::ndarray<>( self.getData(),
-                                  dim,
-                                  sizes.data(),
-                                  nb::find( self ),  // find the Python object associated with `self` and pass it as owner
-                                  strides.data(),
-                                  nb::dtype< ValueType >(),
-                                  dlpack_device().first,
-                                  device_id );
-         },
-         nb::sig( "def __dlpack__(self, **kwargs: typing.Any) -> typing_extensions.CapsuleType" ) )
-      .def_static( "__dlpack_device__", dlpack_device );
-
-   // NOTE: this is needed only because NumPy does not support writable unversioned dlpacks
-   if constexpr( ! std::is_same_v< DeviceType, TNL::Devices::GPU > && ! std::is_enum_v< ValueType > )
-      array.def(
-         "as_numpy",
-         []( ArrayType& self )
-         {
-            constexpr std::size_t dim = ArrayType::getDimension();
-            std::array< std::size_t, dim > sizes;
-            std::array< std::int64_t, dim > strides;
-            TNL::Algorithms::staticFor< std::size_t, 0, dim >(
-               [ & ]( auto i )
-               {
-                  sizes[ i ] = self.template getSize< i >();
-                  strides[ i ] = self.template getStride< i >();
-               } );
-
-            return nb::ndarray< ValueType, nb::numpy, nb::ndim< dim > >(
-               self.getData(),
-               dim,
-               sizes.data(),
-               nb::find( self ),  // find the Python object associated with `self` and pass it as owner
-               strides.data(),
-               nb::dtype< ValueType >(),
-               nb::device::cpu::value,
-               0  // device_id
-            );
-         },
-         nb::rv_policy::reference_internal,
-         nb::sig( "def as_numpy(self) -> numpy.typing.NDArray[typing.Any]" ),
-         "Returns a NumPy ndarray for this NDArray with shared memory (i.e. the data is not copied)" );
-
-   ndarray_indexing( array );
-   ndarray_iteration( array );
+      //
+      ;
+   }
 }
